from importlib import resources
from typing import List, Optional

import yaml

from v2share.data import V2Data


class ClashConfig:
    def __init__(self, template_path: Optional[str] = None):
        if not template_path:
            template_path = resources.files("v2share.templates") / "clash.yml"
        with open(template_path, "r") as f:
            self.template_data = f.read()
        self.data = {
            "proxies": [],
            "proxy-groups": [],
            "rules": [],
        }
        self.proxy_remarks = []

    def add_proxies(self, proxies: List[V2Data]):
        for proxy in proxies:
            self._add_node(proxy)

    def render(self):
        result = yaml.safe_load(self.template_data)
        result["proxies"] = self.data["proxies"]
        result["rules"] = self.data["rules"]
        result["proxy-groups"][0]["proxies"] = self.proxy_remarks
        return yaml.safe_dump(result)

    def _remark_validation(self, remark, depth: int = 0):
        if remark not in self.proxy_remarks:
            return remark
        return self._remark_validation(f"{remark} {depth}", depth + 1)

    def _make_node(
        self,
        name: str,
        type: str,
        server: str,
        port: int,
        network: str,
        tls: bool,
        sni: str,
        host: str,
        path: str,
        udp: bool = True,
        alpn: str = "",
        ais: bool = "",
    ):
        if type == "shadowsocks":
            type = "ss"

        remark = self._remark_validation(name)
        node = {
            "name": remark,
            "type": type,
            "server": server,
            "port": port,
            "network": network,
            f"{network}-opts": {},
            "udp": udp,
        }

        if type == "ss":  # shadowsocks
            return node

        if tls:
            node["tls"] = True
            if type == "trojan":
                node["sni"] = sni
            else:
                node["servername"] = sni
            if alpn:
                node["alpn"] = alpn.split(",")
            if ais:
                node["skip-cert-verify"] = ais

        net_opts = node[f"{network}-opts"]

        if network == "ws":
            if path:
                net_opts["path"] = path
            if host:
                net_opts["headers"] = {"Host": host}

        if network == "grpc":
            if path:
                net_opts["grpc-service-name"] = path

        if network == "h2":
            if path:
                net_opts["path"] = path
            if host:
                net_opts["host"] = [host]

        if network == "http" or network == "tcp":
            if path:
                net_opts["method"] = "GET"
                net_opts["path"] = [path]
            if host:
                net_opts["method"] = "GET"
                net_opts["headers"] = {"Host": host}

        return node

    def _add_node(self, config: V2Data):
        node = self._make_node(
            name=config.remark,
            type=config.protocol,
            server=config.address,
            port=config.port,
            network=config.transport_type,
            tls=(config.tls == "tls"),
            sni=config.sni,
            host=config.host,
            path=config.path,
            udp=True,
            alpn=config.alpn,
            ais=config.allow_insecure,
        )

        if config.protocol == "vmess":
            node["uuid"] = str(config.uuid)
            node["alterId"] = 0
            node["cipher"] = "auto"
            self.data["proxies"].append(node)
            self.proxy_remarks.append(config.remark)
        elif config.protocol == "trojan":
            node["password"] = config.password
            self.data["proxies"].append(node)
            self.proxy_remarks.append(config.remark)
        elif config.protocol == "shadowsocks":
            node["password"] = config.password
            node["cipher"] = config.shadowsocks_method
            self.data["proxies"].append(node)
            self.proxy_remarks.append(config.remark)



import base64
import json
import urllib.parse as urlparse
from dataclasses import dataclass
from typing import Optional
from uuid import UUID


@dataclass
class V2Data:
    protocol: str
    remark: str
    address: str
    port: int
    shadowsocks_method: Optional[str] = None
    uuid: Optional[UUID] = None
    vmess_security: str = "auto"
    password: Optional[str] = None
    host: Optional[str] = None
    transport_type: Optional[str] = "tcp"
    grpc_multi_mode: bool = False
    path: Optional[str] = None
    kcp_header_type: Optional[str] = None
    header_type: Optional[str] = None
    tls: Optional[str] = "none"
    flow: Optional[str] = None
    sni: Optional[str] = None
    fingerprint: Optional[str] = None
    alpn: Optional[str] = None
    reality_pbk: Optional[str] = None
    reality_sid: Optional[str] = None
    reality_spx: Optional[str] = None
    fragment: bool = False
    fragment_packets: Optional[str] = "tlshello"
    fragment_length: Optional[str] = "100-200"
    fragment_interval: Optional[str] = "10-20"
    allow_insecure: bool = False

    def _apply_tls_settings(self, payload):
        if self.tls == "tls" or self.tls == "reality":
            payload.update(
                {
                    "sni": self.sni,
                    "fp": self.fingerprint,
                }
            )
        if self.tls == "tls":
            payload.update(
                {
                    "alpn": self.alpn,
                }
            )
        if self.tls == "reality":
            payload.update(
                {
                    "pbk": self.reality_pbk,
                    "sid": self.reality_sid,
                    "spx": self.reality_spx,
                }
            )

    def to_link(self):
        if self.protocol == "shadowsocks":
            return (
                "ss://"
                + base64.b64encode(
                    (
                        f"{self.shadowsocks_method}:{self.password}"
                        + f"@{self.address}:{self.port}"
                    ).encode()
                ).decode()
                + f"#{urlparse.quote(self.remark)}"
            )
        elif self.protocol == "vmess":
            payload = {
                "add": self.address,
                "aid": "0",
                "host": self.host,
                "id": str(self.uuid),
                "net": self.transport_type,
                "path": self.path,
                "port": self.port,
                "ps": self.remark,
                "scy": "auto",
                "tls": self.tls,
                "type": self.header_type,
                "v": "2",
            }

            self._apply_tls_settings(payload)
            payload = dict(filter(lambda p: p[1], payload.items()))
            return (
                "vmess://"
                + base64.b64encode(
                    json.dumps(payload, sort_keys=True).encode("utf-8")
                ).decode()
            )
        elif self.protocol == "vless":
            payload = {
                "security": self.tls,
                "type": self.transport_type,
                "host": self.host,
                "headerType": self.header_type,
            }
            if self.flow:
                payload.update({"flow": self.flow})

            path_name = "serviceName" if self.transport_type == "grpc" else "path"
            payload.update({path_name: self.path})

            self._apply_tls_settings(payload)
            payload = dict(filter(lambda p: p[1], payload.items()))
            return (
                "vless://"
                + f"{self.uuid}@{self.address}:{self.port}?"
                + urlparse.urlencode(payload)
                + f"#{(urlparse.quote(self.remark))}"
            )
        elif self.protocol == "trojan":
            payload = {
                "security": self.tls,
                "type": self.transport_type,
                "host": self.host,
                "headerType": self.header_type,
            }
            path_name = "serviceName" if self.transport_type == "grpc" else "path"
            payload.update({path_name: self.path})
            self._apply_tls_settings(payload)
            payload = dict(filter(lambda p: p[1], payload.items()))
            return (
                "trojan://"
                + f"{urlparse.quote(self.password, safe=':')}@{self.address}:{self.port}?"
                + urlparse.urlencode(payload)
                + f"#{urlparse.quote(self.remark)}"
            )



from .clash import ClashConfig
from .clashmeta import ClashMetaConfig
from .data import V2Data
from .singbox import SingBoxConfig
from .xray import XrayConfig

__all__ = [V2Data, SingBoxConfig, ClashConfig, ClashMetaConfig, XrayConfig]



import json
from importlib import resources
from typing import List

from v2share.data import V2Data


class SingBoxConfig(str):
    def __init__(self, template_path: str = None):
        if not template_path:
            template_path = resources.files("v2share.templates") / "singbox.json"
        with open(template_path) as f:
            self._template_data = f.read()
        self._outbounds = []

    def render(self):
        result = json.loads(self._template_data)
        result["outbounds"].extend(self._outbounds)
        urltest_types = ["vmess", "vless", "trojan", "shadowsocks"]
        urltest_tags = [
            outbound["tag"]
            for outbound in result["outbounds"]
            if outbound["type"] in urltest_types
        ]
        selector_types = ["vmess", "vless", "trojan", "shadowsocks", "urltest"]
        selector_tags = [
            outbound["tag"]
            for outbound in result["outbounds"]
            if outbound["type"] in selector_types
        ]

        for outbound in result["outbounds"]:
            if outbound.get("type") == "urltest":
                outbound["outbounds"] = urltest_tags

        for outbound in result["outbounds"]:
            if outbound.get("type") == "selector":
                outbound["outbounds"] = selector_tags

        return json.dumps(result, indent=4)

    @staticmethod
    def tls_config(
        sni=None, fp=None, tls=None, pbk=None, sid=None, alpn=None, ais=None
    ):
        config = {}
        if tls in ["tls", "reality"]:
            config["enabled"] = True

        if sni is not None:
            config["server_name"] = sni

        if tls == "tls" and ais:
            config["insecure"] = ais

        if tls == "reality":
            config["reality"] = {"enabled": True}
            if pbk:
                config["reality"]["public_key"] = pbk
            if sid:
                config["reality"]["short_id"] = sid

        if fp:
            config["utls"] = {"enabled": True, "fingerprint": fp}

        if alpn:
            config["alpn"] = [alpn] if not isinstance(alpn, list) else alpn

        return config

    @staticmethod
    def transport_config(
        transport_type="",
        host="",
        path="",
        method="",
        headers={},
        idle_timeout="15s",
        ping_timeout="15s",
        permit_without_stream=False,
    ):
        transport_config = {}

        if transport_type:
            transport_config["type"] = transport_type

            if transport_type == "http":
                if host:
                    transport_config["host"] = host
                if path:
                    transport_config["path"] = path
                if method:
                    transport_config["method"] = method
                if headers:
                    transport_config["headers"] = headers
                if idle_timeout:
                    transport_config["idle_timeout"] = idle_timeout
                if ping_timeout:
                    transport_config["ping_timeout"] = ping_timeout

            elif transport_type == "ws":
                if path:
                    if "?ed=" in path:
                        path, max_early_data = path.split("?ed=")
                        max_early_data = int(max_early_data)
                        transport_config["early_data_header_name"] = (
                            "Sec-WebSocket-Protocol"
                        )
                        transport_config["max_early_data"] = max_early_data
                    transport_config["path"] = path
                if host:
                    headers.update({"Host": host})
                if headers:
                    transport_config["headers"] = headers

            elif transport_type == "grpc":
                if path:
                    transport_config["service_name"] = path
                if idle_timeout:
                    transport_config["idle_timeout"] = idle_timeout
                if ping_timeout:
                    transport_config["ping_timeout"] = ping_timeout
                if permit_without_stream:
                    transport_config["permit_without_stream"] = permit_without_stream

            elif transport_type == "httpupgrade":
                if host:
                    transport_config["host"] = host
                if path:
                    transport_config["path"] = path
                if headers:
                    transport_config["headers"] = headers
        return transport_config

    @staticmethod
    def make_outbound(
        type: str,
        remark: str,
        address: str,
        port: int,
        net="ws",
        path="",
        host="",
        flow="",
        tls="",
        sni="",
        fp="",
        alpn="",
        pbk="",
        sid="",
        header_type="",
        headers={},
        ais=False,
    ):
        config = {
            "type": type,
            "tag": remark,
            "server": address,
            "server_port": port,
        }
        if flow and type == "vless":
            config["flow"] = flow

        if net in ["http", "ws", "quic", "grpc", "httpupgrade"]:
            config["transport"] = SingBoxConfig.transport_config(
                transport_type=net, host=host, path=path, headers=headers
            )
        else:
            config["network"] = net

        if tls in ("tls", "reality"):
            config["tls"] = SingBoxConfig.tls_config(
                sni=sni, fp=fp, tls=tls, pbk=pbk, sid=sid, alpn=alpn, ais=ais
            )

        return config

    def add_proxies(self, proxies: List[V2Data]):
        for config in proxies:
            outbound = self.make_outbound(
                remark=config.remark,
                type=config.protocol,
                address=config.address,
                port=config.port,
                net=config.transport_type,
                tls=config.tls,
                flow=config.flow,
                sni=config.sni,
                host=config.host,
                path=config.path,
                alpn=config.alpn,
                fp=config.fingerprint,
                pbk=config.reality_pbk,
                sid=config.reality_sid,
                header_type=config.header_type,
                headers={},
                ais=config.allow_insecure,
            )

            if config.protocol in ["vless", "vmess"]:
                outbound["uuid"] = str(config.uuid)

            elif config.protocol == "trojan":
                outbound["password"] = config.password

            elif config.protocol == "shadowsocks":
                outbound["password"] = config.password
                outbound["method"] = config.shadowsocks_method

            self._outbounds.append(outbound)



from v2share.clash import ClashConfig
from v2share.data import V2Data


class ClashMetaConfig(ClashConfig):
    def _make_node(
        self,
        name: str,
        type: str,
        server: str,
        port: int,
        network: str,
        tls: bool,
        sni: str,
        host: str,
        path: str,
        header_type: str = "",
        udp: bool = True,
        alpn: str = "",
        fp: str = "",
        pbk: str = "",
        sid: str = "",
        ais: bool = "",
    ):
        node = super()._make_node(
            name=name,
            type=type,
            server=server,
            port=port,
            network=network,
            tls=tls,
            sni=sni,
            host=host,
            path=path,
            udp=udp,
            alpn=alpn,
            ais=ais,
        )
        if fp:
            node["client-fingerprint"] = fp
        if pbk:
            node["reality-opts"] = {"public-key": pbk, "short-id": sid}

        return node

    def _add_node(self, config: V2Data):
        node = self._make_node(
            name=config.remark,
            type=config.protocol,
            server=config.address,
            port=config.port,
            network=config.transport_type,
            tls=(config.tls in ["tls", "reality"]),
            sni=config.sni,
            host=config.host,
            path=config.path,
            header_type=config.header_type,
            udp=True,
            alpn=config.alpn,
            fp=config.fingerprint,
            pbk=config.reality_pbk,
            sid=config.reality_sid,
            ais=config.allow_insecure,
        )

        if config.protocol == "vmess":
            node["uuid"] = str(config.uuid)
            node["alterId"] = 0
            node["cipher"] = "auto"
            self.data["proxies"].append(node)
            self.proxy_remarks.append(config.remark)

        if config.protocol == "vless":
            node["uuid"] = str(config.uuid)

            if config.transport_type in ("tcp", "kcp") and config.header_type != "http":
                node["flow"] = config.flow

            self.data["proxies"].append(node)
            self.proxy_remarks.append(config.remark)

        if config.protocol == "trojan":
            node["password"] = config.password

            if (
                config.transport_type in ("tcp", "kcp")
                and config.header_type != "http"
                and config.tls
            ):
                node["flow"] = config.flow

            self.data["proxies"].append(node)
            self.proxy_remarks.append(config.remark)

        if config.protocol == "shadowsocks":
            node["password"] = config.password
            node["cipher"] = config.shadowsocks_method
            self.data["proxies"].append(node)
            self.proxy_remarks.append(config.remark)



import json
from importlib import resources
from typing import List

from v2share.data import V2Data


class XrayConfig(str):

    def __init__(self, template_path: str = None, mux_template_path: str = None):
        self.config = []
        if not template_path:
            template_path = resources.files("v2share.templates") / "xray.json"
        if not mux_template_path:
            mux_template_path = resources.files("v2share.templates") / "xray_mux.json"
        with open(template_path) as f:
            self._template = f.read()
        with open(mux_template_path) as f:
            self._mux_template = f.read()

    def add_proxies(self, proxies: List[V2Data]):
        for data in proxies:
            outbound = {"tag": data.remark, "protocol": data.protocol}

            if data.protocol == "vmess":
                outbound["settings"] = XrayConfig.vmess_config(
                    address=data.address, port=data.port, uuid=str(data.uuid)
                )
            elif data.protocol == "vless":
                if data.tls == "reality":
                    flow = data.flow
                else:
                    flow = None
                outbound["settings"] = XrayConfig.vless_config(
                    address=data.address, port=data.port, uuid=str(data.uuid), flow=flow
                )

            elif data.protocol == "trojan":
                outbound["settings"] = XrayConfig.trojan_config(
                    address=data.address, port=data.port, password=data.password
                )

            elif data.protocol == "shadowsocks":
                outbound["settings"] = XrayConfig.shadowsocks_config(
                    address=data.address,
                    port=data.port,
                    password=data.password,
                    method=data.shadowsocks_method,
                )

            outbounds = [outbound]
            dialer_proxy = None

            if data.fragment:
                fragment_outbound = XrayConfig.make_fragment_outbound(
                    data.fragment_packets, data.fragment_length, data.fragment_interval
                )
                outbounds.append(fragment_outbound)
                dialer_proxy = fragment_outbound["tag"]

            outbound["streamSettings"] = XrayConfig.make_stream_settings(
                net=data.transport_type,
                tls=data.tls,
                sni=data.sni,
                host=data.host,
                path=data.path,
                alpn=data.alpn,
                fp=data.fingerprint,
                pbk=data.reality_pbk,
                sid=data.reality_sid,
                ais=data.allow_insecure,
                kcp_header_type=data.kcp_header_type,
                grpc_multi_mode=data.grpc_multi_mode,
                dialer_proxy=dialer_proxy,
            )

            mux_config = json.loads(self._mux_template)

            outbound["mux"] = mux_config
            if outbound["mux"]["enabled"]:
                outbound["mux"]["enabled"] = False

            json_template = json.loads(self._template)
            json_template["remarks"] = data.remark
            json_template["outbounds"] = outbounds + json_template["outbounds"]
            self.config.append(json_template)

    def render(self):
        return json.dumps(self.config, indent=4)

    @staticmethod
    def tls_config(sni, fingerprint, alpn=None, ais=False):
        tlsSettings = {"serverName": sni, "allowInsecure": ais or False}
        if alpn is None:
            alpn = ["h2", "http/1.1"]

        tlsSettings["alpn"] = alpn

        if fingerprint:
            tlsSettings["fingerprint"] = fingerprint

        return tlsSettings

    @staticmethod
    def reality_config(public_key, short_id, sni, fingerprint="", spiderx=""):

        realitySettings = {
            "serverName": sni,
            "fingerprint": fingerprint,
            "show": False,
            "publicKey": public_key,
            "shortId": short_id,
            "spiderX": spiderx,
        }

        return realitySettings

    @staticmethod
    def ws_config(path=None, host=None):

        ws_settings = {}
        if path:
            ws_settings["path"] = path
        if host:
            ws_settings["host"] = host

        return ws_settings

    @staticmethod
    def httpupgrade_config(path=None, host=None):

        httpupgrade_settings = {}
        if path:
            httpupgrade_settings["path"] = path
        if host:
            httpupgrade_settings["host"] = host

        return httpupgrade_settings

    @staticmethod
    def grpc_config(service_name=None, multi_mode=False):

        grpc_settings = {
            "multi_mode": multi_mode,
            "idle_timeout": 60,
            "health_check_timeout": 20,
            "permit_without_stream": False,
            "initial_windows_size": 0,
        }
        if service_name:
            grpc_settings["serviceName"] = service_name
        return grpc_settings

    @staticmethod
    def tcp_http_config(path=None, host=None):
        tcp_settings = {}

        if any((path, host)):
            tcp_settings = {
                "header": {
                    "type": "http",
                    "request": {
                        "version": "1.1",
                        "method": "GET",
                        "headers": {
                            "User-Agent": [],
                            "Accept-Encoding": ["gzip, deflate"],
                            "Connection": ["keep-alive"],
                            "Pragma": "no-cache",
                        },
                    },
                }
            }

            if path:
                tcp_settings["header"]["request"]["path"] = [path]

            if host:
                tcp_settings["header"]["request"]["headers"]["Host"] = [host]

        return tcp_settings

    @staticmethod
    def h2_config(path=None, host=None):
        if host is None:
            host = []
        if path is None:
            path = "/"
        http_settings = {"path": path}
        if host:
            http_settings["host"] = host
        else:
            http_settings["host"] = []

        return http_settings

    @staticmethod
    def quic_config(security="none", key=None, header_type="none"):

        quic_settings = {
            "security": security,
            "header": {"type": header_type},
        }

        if key:
            quic_settings["key"] = key

        return quic_settings

    @staticmethod
    def kcp_config(seed=None, header_type=None):

        kcp_settings = {
            "header": {},
            "mtu": 1350,
            "tti": 50,
            "uplinkCapacity": 12,
            "downlinkCapacity": 100,
            "congestion": False,
            "readBufferSize": 2,
            "writeBufferSize": 2,
        }

        if seed:
            kcp_settings["seed"] = seed
        if header_type:
            kcp_settings["header"]["type"] = header_type
        else:
            kcp_settings["header"]["type"] = "none"

        return kcp_settings

    @staticmethod
    def make_stream_settings(
        net="tcp",
        path=None,
        host=None,
        tls="none",
        sni="",
        fp=None,
        alpn="h2,http/1.1",
        pbk=None,
        sid=None,
        ais=False,
        kcp_header_type=None,
        grpc_multi_mode=False,
        dialer_proxy=None,
    ):
        stream_settings = {"network": net}

        if net == "ws":
            stream_settings["wsSettings"] = XrayConfig.ws_config(path=path, host=host)
        elif net == "grpc":
            stream_settings["grpcSettings"] = XrayConfig.grpc_config(
                service_name=path, multi_mode=grpc_multi_mode
            )
        elif net == "h2":
            stream_settings["httpSettings"] = XrayConfig.h2_config(path=path, host=host)
        elif net == "kcp":
            stream_settings["kcpSettings"] = XrayConfig.kcp_config(
                seed=path, header_type=kcp_header_type
            )
        elif net == "tcp":
            stream_settings["tcpSettings"] = XrayConfig.tcp_http_config(
                path=path, host=host
            )
        elif net == "quic":
            stream_settings["quicSettings"] = XrayConfig.quic_config()
        elif net == "httpupgrade":
            stream_settings["httpupgradeSettings"] = XrayConfig.httpupgrade_config(
                path=path, host=host
            )

        if tls == "tls":
            stream_settings["security"] = "tls"
            stream_settings["tlsSettings"] = XrayConfig.tls_config(
                sni=sni, fingerprint=fp, alpn=alpn.split(",") if alpn else None, ais=ais
            )
        elif tls == "reality":
            stream_settings["security"] = "xtls"
            stream_settings["realitySettings"] = XrayConfig.reality_config(
                pbk, sid, sni, fingerprint=fp
            )
        else:
            stream_settings["security"] = "none"

        if dialer_proxy:
            stream_settings["sockopt"] = {"dialerProxy": dialer_proxy}

        return stream_settings

    @staticmethod
    def vmess_config(address, port, uuid):

        return {
            "vnext": [
                {
                    "address": address,
                    "port": port,
                    "users": [
                        {
                            "id": uuid,
                            "security": "auto",
                        }
                    ],
                }
            ]
        }

    @staticmethod
    def vless_config(address, port, uuid, flow="none"):

        return {
            "vnext": [
                {
                    "address": address,
                    "port": port,
                    "users": [
                        {
                            "id": uuid,
                            "encryption": "none",
                            "flow": flow,
                        }
                    ],
                }
            ]
        }

    @staticmethod
    def trojan_config(address, port, password):
        return {
            "servers": [
                {
                    "address": address,
                    "port": port,
                    "password": password,
                    "email": "kiomars",
                }
            ]
        }

    @staticmethod
    def shadowsocks_config(address, port, password, method):
        return {
            "servers": [
                {
                    "address": address,
                    "port": port,
                    "password": password,
                    "email": "kiomars",
                    "method": method,
                    "uot": False,
                }
            ]
        }

    @staticmethod
    def make_fragment_outbound(packets, length, interval):
        return {
            "tag": "fragment_out",
            "protocol": "freedom",
            "settings": {
                "fragment": {"packets": packets, "length": length, "interval": interval}
            },
        }
